//---------------------------------------------------------------------
// <copyright file="EFOracleVersion.cs" company="Microsoft">
//      Copyright (c) Microsoft Corporation.  All rights reserved.
// </copyright>
//---------------------------------------------------------------------

namespace EFOracleProvider
{
    using System;

    /// <summary>
    /// This enum describes the current storage version
    /// </summary>
    internal enum EFOracleVersion
    {
        /// <summary>
        /// Oracle9i
        /// </summary>
        Oracle9i = 9,

        /// <summary>
        /// Oracle10g
        /// </summary>
        Oracle10g = 10,

        /// <summary>
        /// Oracle 11g
        /// </summary>
        Oracle11g = 11,

        // higher versions go here
    }

    /// <summary>
    /// This class is a simple utility class that determines the version from the 
    /// connection
    /// </summary>
    internal static class EFOracleVersionUtils
    {
        /// <summary>
        /// Get the version from the connection.
        /// </summary>
        /// <param name="connection">current connection</param>
        /// <returns>version for the current connection</returns>
        internal static EFOracleVersion GetStorageVersion(EFOracleConnection connection)
        {
            string serverVersion = connection.ServerVersion;
            if (serverVersion.StartsWith("9."))
            {
                return EFOracleVersion.Oracle9i;
            }
            else if (serverVersion.StartsWith("10."))
            {
                return EFOracleVersion.Oracle10g;
            }
            else if (serverVersion.StartsWith("11."))
            {
                return EFOracleVersion.Oracle11g;
            }

            throw new ArgumentException("Could not determine storage version; " +
                    "a valid storage connection or a version hint is required.");
        }

        internal static string GetVersionHint(EFOracleVersion version)
        {
            switch (version)
            {
                case EFOracleVersion.Oracle9i:
                    return EFOracleProviderManifest.TokenOracle9i;

                case EFOracleVersion.Oracle10g:
                    return EFOracleProviderManifest.TokenOracle10g;

                case EFOracleVersion.Oracle11g:
                    return EFOracleProviderManifest.TokenOracle11g;

                default:
                    throw new ArgumentException("Could not determine storage version; " +
                            "a valid storage connection or a version hint is required.");
            }
        }

        internal static EFOracleVersion GetStorageVersion(string versionHint)
        {
            if (!string.IsNullOrEmpty(versionHint))
            {
                switch (versionHint)
                {
                    case EFOracleProviderManifest.TokenOracle9i:
                        return EFOracleVersion.Oracle9i;

                    case EFOracleProviderManifest.TokenOracle10g:
                        return EFOracleVersion.Oracle10g;

                    case EFOracleProviderManifest.TokenOracle11g:
                        return EFOracleVersion.Oracle11g;
                }
            }

            throw new ArgumentException("Could not determine storage version; " +
                    "a valid storage connection or a version hint is required.");
        }

        internal static bool IsVersionX(EFOracleVersion storageVersion)
        {
            return storageVersion == EFOracleVersion.Oracle9i || storageVersion == EFOracleVersion.Oracle10g ||
                storageVersion == EFOracleVersion.Oracle11g;
        }
    }
}
